/*
 * DeltaEditorData.java
 *
 * Created on 11. Februar 2004, 14:01
 */

package net.metadiversity.diversity.navigator.db.delta_editor;

import org.apache.log4j.Logger;
import net.metadiversity.diversity.navigator.gui.DnOptionPane;
import net.metadiversity.diversity.navikey.util.Delta_pre_selectionScanner;
import net.metadiversity.diversity.navigator.util.DiversityAcceptedNames;
import net.metadiversity.diversity.navigator.util.FormatUnit;
/**
 *
 * @author  didi
 */
public class DeltaEditorData 
{
    static Logger logger = Logger.getLogger( DeltaEditorData.class.getName() ); 
    
    private int characterCount;
    private int itemNameCount;
    
    private boolean showItemIndex   = false;
    private boolean showHeaderIndex = false;
        
    boolean debug = false;
    
    // private String TABLE_PREFIX = "" + TABLE_PREFIX + "";
    private String TABLE_PREFIX = "LIASlight";
    
    
    java.util.Vector charNameVector     = null;
    java.util.Vector itemNameVector     = null;
    java.util.Vector inapplicableVector = null;
    java.util.Vector dependenciesVector = null;
    
    java.sql.Connection con             = null;
    
    java.sql.PreparedStatement getCharacterDataStatment     = null;
    java.sql.PreparedStatement getCharacterStatesStatment   = null;
    
    java.sql.Statement getTableElementStatement 			= null;

    
    private String taxonNamesSql;
    private String characterNamesSql;
    
    String tableCHAR    = null;
    String tableITEM    = null;
    String tableDESCR   = null;
    String tableCS      = null;
    String tableDEP     = null;
    
    public static final int    DESCR_CS         = 1;
    public static final int    DESCR_MODIFIER   = 2;
    public static final int    DESCR_X          = 3;
    public static final int    DESCR_TXT        = 4;
    public static final int    DESCR_NOTES      = 5;
    public static final int    DESCR_SEQ        = 6;
    
    // NaviKey Data
    private    String specs = "";
    private    String chars = "";
    private    String items = "";
    
    // AcceptedNames
    private DiversityAcceptedNames diversityAcceptedNames = null;
    private boolean showAcceptedNamesFlag;        
            
    
    /** Creates a new instance of DeltaEditorData */
    public DeltaEditorData( java.sql.Connection con, java.sql.Connection connectionAcceptedNames, String tablePrefix ) 
    {        
        TABLE_PREFIX = tablePrefix;
        this.con = con;
        
        // init Accepted Names
        if( connectionAcceptedNames != null )
        {
            diversityAcceptedNames = new DiversityAcceptedNames( connectionAcceptedNames ); 
        }
        // init
        tableCHAR = "\"" + TABLE_PREFIX + "_CHAR\"";
        characterNamesSql = "SELECT \"CID\", \"CharName\", \"Unit\", \"Type\" FROM " + tableCHAR + " ORDER BY \"CID\"; ";

        tableITEM = "\"" + TABLE_PREFIX + "_ITEM\"";
        if( TABLE_PREFIX.startsWith( "BIOTA") )
        {
            // no "NameID"
            taxonNamesSql    = "SELECT \"IID\", \"ItemName\" FROM " + tableITEM + " ORDER BY \"ItemName\"; ";
        }
        else
        {
            taxonNamesSql    = "SELECT \"IID\", \"ItemName\", \"NameID\" FROM " + tableITEM + " ORDER BY \"ItemName\"; ";
        }
                
        tableDESCR = "\"" + TABLE_PREFIX + "_DESCR\"";
        String taxonDescriptionSql = "SELECT \"CS\",\"Modifier\", \"X\", \"TXT\", \"Notes\", \"SEQ\" FROM " + tableDESCR + " WHERE \"IID\"=? AND \"CID\"=? ORDER BY \"CS\" ASC";
        
        tableCS = "\"" + TABLE_PREFIX + "_CS\"";
        tableDEP = "\"" + TABLE_PREFIX + "_DEP\"";
        
        String characterStatesSql = "SELECT \"CS\", \"CharStateName\" FROM " + tableCS + " WHERE \"CID\"=?  ORDER BY  \"StateID\"; ";
                
        
        charNameVector      = new java.util.Vector();
        inapplicableVector  = new java.util.Vector();
        dependenciesVector           = new java.util.Vector();
        try
        {
            java.sql.Statement stmt = con.createStatement();
            //------------------------
            // init PreparedStatement 
            //------------------------
            getCharacterDataStatment    = con.prepareStatement( taxonDescriptionSql );
            getCharacterStatesStatment  = con.prepareStatement( characterStatesSql );
            
            getTableElementStatement 	= con.createStatement();

            // setup CharacterNames
            buildCharacterNameVector();
            // setup ItemNames
            buildItemNameVector();
            // setup Inapplicable
            buildInapplicableVector();
        }
        catch( java.sql.SQLException ex )
        {
            logger.error( ex.getMessage() );             
        }        
        // set size of row and columne
        characterCount = charNameVector.size();
        itemNameCount = itemNameVector.size();
        // init data for inapplicalible test
        initInapplicalibleVector();
    }   
    //---------------------------------------------------------------------------
    // build itemNameVector
    //--------------------------------------------------------------------------
    private void buildItemNameVector()
    {
        try
        {
            java.sql.Statement stmt = con.createStatement();
            java.sql.ResultSet rs;
            int index;
            //-------------
            // item names
            //-------------
            rs   = stmt.executeQuery( taxonNamesSql );
            index = 0;
            itemNameVector = new java.util.Vector();
            while( rs.next() )
            {
                int     iid             = rs.getInt( "IID" );
                String  itemName        = rs.getString( "ItemName" );
                
                int     nameId;
                if( TABLE_PREFIX.startsWith( "BIOTA") )
                {
                    nameId          = rs.getInt( "IID" );
                }
                else
                {
                    nameId          = rs.getInt( "NameID" );                    
                }
                
                if( diversityAcceptedNames != null && showAcceptedNamesFlag )
                {
                    final String mark = "(A) ";
                    boolean showMark = false;
                    
                    itemName = showMark ? mark : "" + diversityAcceptedNames.getAcceptedName( nameId );
                }                
                itemNameVector.add( index++, new ItemElement( iid,  itemName, nameId ) );
            }
            stmt.close();
            itemNameCount = itemNameVector.size();
            // logger.error( "ItemNameCount: " + itemNameCount + " Index: " + index );             
        }
        catch( java.sql.SQLException ex )
        {
            logger.error( ex.getMessage() );             
        }                
    }
    //--------------------------------------------------------------------------
    //
    //--------------------------------------------------------------------------
    private void buildCharacterNameVector()
    {
        try
        {
            java.sql.Statement stmt = con.createStatement();
            java.sql.ResultSet rs;
            int index = 0;
            //------------------
            // Character names
            //------------------
            rs   = stmt.executeQuery( characterNamesSql );
            while( rs.next() )
            {
                int     cid             = rs.getInt( "CID" );
                String  characterName   = rs.getString( "CharName" );
                String  unit            = rs.getString( "Unit" );
                String  type            = rs.getString( "Type" );
                
                charNameVector.add( index++, new HeadElement( cid,  characterName, unit, type ) );
            }
            stmt.close();
        }
        catch( java.sql.SQLException ex )
        {
            logger.error( ex.getMessage() );             
        }        
        
    }
    //--------------------------------------------------------------------------
    //
    //--------------------------------------------------------------------------
    private void buildInapplicableVector()
    {
        try
        {
            java.sql.Statement stmt = con.createStatement();
            java.sql.ResultSet rs;
            int index = 0;
            //-----------------
            // inapplicalible
            //-----------------
            rs   = stmt.executeQuery( "SELECT DISTINCT \"InapplicableCID\" FROM " + tableDEP );
            while( rs.next() )
            {
                String  inApp = rs.getString( 1 );                
                inapplicableVector.add( inApp );
            }
            stmt.close();
        }
        catch( java.sql.SQLException ex )
        {
            logger.error( ex.getMessage() );             
        }        
        
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public int getItemNameCount()
    {
        return itemNameCount;
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public int getCharacterCount()
    {
        return characterCount;
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public void showItemIndex( boolean b )
    {
        showItemIndex = b;
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public void showHeaderIndex( boolean b )
    {
        showHeaderIndex = b;
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public int getIID( int row )
    {
        return (( ItemElement ) itemNameVector.get( row )).getIID(); // offest for HeadElement
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public int getNameID( int row )
    {
        if( row < 0 )
        {
            return -1;  // ???
        }
        return (( ItemElement ) itemNameVector.get( row )).getNameId(); // offest for HeadElement
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public int getCID( int col )
    {
        return (( HeadElement ) charNameVector.get( col )).getCID(); // offset for ItemElement
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public CharacterStateDescription getTableElement( int row, int col )
    {
        // logger.debug( "getTableElement( " + row + ", " + col + ")" );
        // System.out.println( "getTableElement( " + row + ", " + col + ")" ); // ???
        HeadElement hl = ( HeadElement ) charNameVector.get( col ); // offset for ItemElement
        ItemElement ie;
                
        try
        {
            ie = ( ItemElement ) itemNameVector.get( row ); // offest for HeadElement
        }
        catch( java.lang.Exception ex )
        {
            return null;
        }
        int iid = ie.getIID();
        int cid = hl.getCID();
        
        java.lang.String characterType = hl.getType();
        
        boolean isInAppl = false; 

        try
        {
            //-------------------------
            // Inapplicabile 
            // ??? Not tested well ??? 
            //-------------------------                       
            if( inapplicableVector.contains( String.valueOf( cid ) ) )
            {   
                isInAppl = true;
      
                InapplicalibleTest it = new InapplicalibleTest( dependenciesVector, cid );
                
                String sql = "SELECT \"CID\", \"CS\" FROM " + tableDESCR + " WHERE \"IID\"=" + iid + " AND ( " + it.getWhereClause() + " )";
                // System.out.println( sql ); // ???
                // logger.debug( sql );
                // java.sql.Statement getTableElementStatement = con.createStatement();
                java.sql.ResultSet irs = getTableElementStatement.executeQuery( sql );
                while( irs.next() )
                {
                    // logger.debug( irs.getString( 1 ) + " " + irs.getString( 2 ) );
                    isInAppl = it.isInapplicalible( irs.getInt( 1 ), irs.getString( 2 ) );
                }
                //isInAppl = it.isInapplicalible();
            }            
            //------------------------
            //  get Character States
            //------------------------
            getCharacterDataStatment.setInt( 1, iid );
            getCharacterDataStatment.setInt( 2, cid );

            java.sql.ResultSet rs = getCharacterDataStatment.executeQuery();

            CharacterStateDescription csd = new CharacterStateDescription();
            csd.setInapplicalible( isInAppl );
            while( rs.next() )
            {
                csd.add( new CharacterStateDescriptionEntry( iid, cid, 
                                                             rs.getString( DESCR_MODIFIER ),
                                                             rs.getString( DESCR_CS ), 
                                                             rs.getDouble( DESCR_X ), 
                                                             rs.getString( DESCR_TXT ), 
                                                             rs.getString( DESCR_NOTES ), 
                                                             rs.getInt( DESCR_SEQ )
                                                             // ??? ,characterType                                                             
                                                           ) 
                        );
            }
            return csd;        
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );             
        }
        logger.fatal( "Software error" );
        return null;        
    }
    //--------------------------------------------------------------------------
    // get IID
    //--------------------------------------------------------------------------    
    public int getItemIID( int row )
    {
        ItemElement ie = ( ItemElement ) itemNameVector.get( row ); // ??? offest for HeadElement
        return ie.getIID();        
    }
    //--------------------------------------------------------------------------
    // get Lable for this Item 
    //--------------------------------------------------------------------------
    public String getItemLableText( int row )
    {    
        // Show Taxon Name
        String str;
        ItemElement ie = ( ItemElement ) itemNameVector.get( row ); // ??? offest for HeadElement
        str = ( String ) ie.getItemName();
        if( showItemIndex )
        {
            boolean showIndex = false;  // false --> show line number

            java.text.NumberFormat nf = java.text.NumberFormat.getInstance();
            nf.setMinimumIntegerDigits( 3 );
            if( showIndex )
            {
                str = nf.format( ie.getIID() ) + " " + str;
            }
            else
            {
                // show line number (start with 1, row starts with 0 )
                int line = row + 1;
                String n = "" + line + "";                    
                while( n.length() < 4 )
                {
                    n = "0" + n;
                }
                str = n  + " " + str;
            }
        }
        return str;
    }
    //--------------------------------------------------------------------------
    //
    //--------------------------------------------------------------------------
    public String getItemText( int row )
    {
        if( debug ) logger.debug( "getItemText( row = " + row + " )" );
            // Show Taxon Name
            String str;
            ItemElement ie = ( ItemElement ) itemNameVector.get( row ); // ??? offest for HeadElement
            str = ( String ) ie.getItemName();
            return str;
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public String getHeadIndex( int col )
    {
        if( debug ) logger.debug( "getHeadIndex( col = " + col + " )" );
        
        if( charNameVector != null )
        {
            HeadElement he = ( HeadElement ) charNameVector.get( col ); // ??? offset for ItemElement
            
            String str = ( String ) he.getCharacterName();
            // str = str.replaceAll( " ", "<p>" );
            if( showHeaderIndex )
            {
                str = "<html>" + he.getCID() + "<p>" + str + "</html>";
            }
            else
            {
                str = "<html>" + str + "</html>";
            }            
            return str;
        }
        else
        {
            return "none";
        }
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public String getHeadText( int col )
    {
        if( debug ) logger.debug( "getHeadText( col = " + col + " )" );
        
        String str;
        HeadElement hl = ( HeadElement ) charNameVector.get( col ); // ??? offset for ItemElement
        str = ( String ) hl.getCharacterName();
        // str = str.replaceAll( " ", "<p>" );
        // str = "<html>" + hl.getCID() + "<p>" + str + "</html>";
        str = "<html>" + str + "</html>";
        return str;
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public String getTypeUnit( int col )
    {
        if( debug ) logger.debug( "getHeadTypUnit( col = " + col + " )" );
        
        String str;
        HeadElement hl = ( HeadElement ) charNameVector.get( col ); // ??? offset for ItemElement
        str = hl.getUnit();
        if( str == null )
        {
            return null;
        }
        str = ( String ) hl.getType() + ": " + str;
        return str;
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    class HeadElement
    {
        int     cid;
        String  characterName;
        String  unit;
        String  type;
        
        public HeadElement( int cid, String characterName, String unit, String type )
        {
            this.cid = cid;
            this.characterName = characterName;
            this.unit = unit;
            this.type = type;
        }
        public int getCID()
        {
            return cid;
        }
        public String getCharacterName()
        {
            return characterName;
        }
        public String getType()
        {
            return type;
        }
        public String getUnit()
        {
            return unit;
        }
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    class ItemElement
    {
        int     iid;
        String  name;
        int		nameId;
        
        public ItemElement( int iid, String itemName, int nameId )
        {
            this.iid = iid;
            this.name = itemName;
            this.nameId = nameId;
        }
        public int getIID()
        {
            return iid;
        }
        public String getItemName()
        {
            return name;
        }
        public int getNameId()
        {
            return nameId;
        }
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------        
    public java.util.Vector getCharacterStatesNames( int col )
    {
        if( debug ) logger.debug( "getCharacterStatesNames( col = " + col + " )" );
        
        HeadElement hl = ( HeadElement ) charNameVector.get( col ); // offset for ItemElement
        int cid = hl.getCID();
        
        if( debug ) logger.debug( "getCharacterStatesNames( cid = " + cid + " )" );
        try
        {
            getCharacterStatesStatment.setInt( 1, cid );
            java.sql.ResultSet rs = getCharacterStatesStatment.executeQuery();

            java.util.Vector result = new java.util.Vector();
            while( rs.next() )
            {
                
                result.add( new CharacterStateElement( rs.getString( 1 ), rs.getString( 2 ) ) );                                
            }                       
            //  Sort: 1,2,11,... Character Stat name
            // not posible to sort by "StateID"
            java.util.Collections.sort( result );
            return result;
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );
            return null;
        }
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------        
    public void initInapplicalibleVector()
    {
        try
        {
            java.sql.Statement stmt = con.createStatement();
            java.sql.ResultSet rs   = stmt.executeQuery( "SELECT DISTINCT \"CID\", \"CS\", \"InapplicableCID\" FROM " + tableDEP + " ORDER BY \"InapplicableCID\", \"CID\", \"CS\" " );
            while( rs.next() )
            {
                int     cid     = rs.getInt( 1 );
                String  cs      = rs.getString( 2 );
                int     dep     = rs.getInt( 3 );                
                
                dependenciesVector.add( new Dependencies( cid, cs, dep ) );
            }
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );             
        }
    }    
    //--------------------------------------------------------------------------
    // Database add/update/delete CharacterDescriptions 
    // Table: TABLE_PREFIX + tableDESCR
    // PrimaryKey: 
    //--------------------------------------------------------------------------
    private String getDescrWhereClause( CharacterStateDescriptionEntry csde )
    {
        return " WHERE (\"IID\"=" + csde.getIID() + " AND \"CID\"=" + csde.getCID() + " AND \"CS\"=" + csde.getCSSqlValue() + " )";
    }
    private String getDescrColumnNames( CharacterStateDescriptionEntry csde )
    {
        return " ( \"IID\",\"CID\",\"CS\",\"Modifier\", \"X\", \"TXT\", \"Notes\", \"SEQ\" ) ";
    }
    private String getDescrValues( CharacterStateDescriptionEntry csde )
    {
        return " VALUES ( " + csde.getIID() + ", " + csde.getCID() + ", "  + csde.getCSSqlValue() + ", " + csde.getModifierSqlValue() + ", " + csde.getX() + ", " + csde.getTextSqlValue() + ", " + csde.getNotesSqlValue() + ", " + csde.getSEQ() + " )";
    }
    private String getDescrSet( CharacterStateDescriptionEntry csde )
    {
        return " SET \"IID\"=" + csde.getIID() + ", \"CID\"=" + csde.getCID() + ", \"CS\"='" + csde.getCS() + "', \"Modifier\"=" + csde.getModifierSqlValue() + ", \"X\"=" + csde.getX() + ", \"TXT\"=" + csde.getTextSqlValue() + ", \"Notes\"=" + csde.getNotesSqlValue() + ", \"SEQ\"=" + csde.getSEQ() + " ";
    }
    //--------------------------------------------------------------------------
    public void inserCharacterStateDescriptionEntry( CharacterStateDescriptionEntry csde )
    {
        // add to datebase
        String sql = "INSERT INTO " + tableDESCR + getDescrColumnNames( csde ) + getDescrValues( csde );
        
        logger.debug( sql );
        executeUpdate( sql );
        // add to data
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public void removeCharacterStateDescriptionEntry( CharacterStateDescriptionEntry csde )
    {
        // remove from datebase
        String sql = "DELETE FROM " + tableDESCR + getDescrWhereClause( csde );
        
        logger.debug( sql );
        executeUpdate( sql );
        // remove data
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public void updateCharacterStateDescriptionEntry( CharacterStateDescriptionEntry csde )
    {
        // update datebase
        String sql = "UPDATE " + tableDESCR + getDescrSet( csde ) + getDescrWhereClause( csde );
        
        logger.debug( sql );
        executeUpdate( sql );
        // update data
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    private void executeUpdate( String sql )
    {
        try
        {
            java.sql.Statement stmt = con.createStatement();
            int rowCount = stmt.executeUpdate( sql );
            if( rowCount != 1 )
            {
                logger.fatal( "Update ERROR : " + sql );
            }
            stmt.close();
        }
        catch( java.sql.SQLException ex )
        {
            DnOptionPane.DnShowMessageDialog( null, ex.getMessage(), "Update error", DnOptionPane.ERROR_MESSAGE );                
            logger.fatal( ex.getMessage() );             
        }        
    }
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    public java.lang.String getCharacterType( int cid )
    {
        java.util.ListIterator iter = charNameVector.listIterator();
        
        while( iter.hasNext() )
        {
            HeadElement hl = ( HeadElement ) iter.next();
            
            if( hl.cid == cid )
            {
                return hl.getType();
            }
        }
        return null;
    }    
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------
    private int getNextItemID()
    {
        try
        {
            java.sql.Statement stmt = con.createStatement();
            java.sql.ResultSet rs   = stmt.executeQuery( "SELECT max( \"IID\" ) FROM " + tableITEM );  
            int iid = -2;
            if( rs.next() )
            {
                iid = rs.getInt( 1 );
            }                
            rs.close();
            stmt.close();            
            return ++iid;
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );             
        }        
        return -1;        
    }
    //--------------------------------------------------------------------------
    // return iid
    //--------------------------------------------------------------------------    
    public int insertNewItem( String itemName, int nameId )
    {
        int iid = getNextItemID();        
        
        // ??? DiversityDescription Version not the same
        // BIOTA: USE "IID" and ItemID"
        // LIASlight: ONLY "IID"
        // LIASlight: Version 19b USE "IID" and ItemID" and "NameID"
        String insertNewItemSql;

        //???????
        
        insertNewItemSql = "INSERT INTO " + tableITEM + "( \"IID\", \"ItemID\", \"ItemName\", \"Abundance\") VALUES ( " + iid + ", " + iid + ", '" + itemName + "', " +  5.0 + " )";
        
        if( tableITEM.startsWith( "\"BIOTA" ) )
        {
            insertNewItemSql = "INSERT INTO " + tableITEM + "( \"IID\", \"ItemID\", \"ItemName\" ) VALUES ( " + iid + ", " + iid + ", '" + itemName + "' )";
        }
        else if( tableITEM.startsWith( "\"LIASlight" ) )
        {
            insertNewItemSql = "INSERT INTO " + tableITEM + "( \"IID\", \"ItemID\", \"ItemName\", \"NameID\" ) VALUES ( " + iid + ", " + iid + ", '" + itemName + "', " + nameId + " )";
        }
        else
        {
            insertNewItemSql = "INSERT INTO " + tableITEM + "( \"IID\", \"ItemName\" ) VALUES ( " + iid + ", '" + itemName + "' )";            
        }
        try
        {
            java.sql.Statement stmt = con.createStatement();
            logger.fatal( "Delta Editor INSERT: " + insertNewItemSql );
            if( stmt.execute( insertNewItemSql ) )
            {
                logger.fatal( "INSERT ERROR : " + insertNewItemSql );
            }
            stmt.close();
            // Update ( because GUI must changed )
            buildItemNameVector();
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );             
        }        
        return iid;
    }    
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------    
    public void deleteItem( int iid, String itemName )
    {
        String deleteItemSql = "DELETE FROM " + tableITEM + " WHERE \"ItemName\" = '" + itemName + "' AND \"IID\" = " + iid;                  
        String selectItemSql = "SELECT COUNT( * ) FROM " + tableITEM + " WHERE \"ItemName\" = '" + itemName + "' ";                  
        try
        {
            java.sql.Statement stmt = con.createStatement();
            java.sql.ResultSet rs = stmt.executeQuery( selectItemSql );
            if( rs == null )
            {
                logger.fatal( "DELETE ERROR : " + selectItemSql );
            }
            else
            {   
                rs.next();
                int count = rs.getInt( 1 );
                if( count > 1 )
                {
                    // ??? 
                    logger.error( "Can't delete item: " + itemName );
                    logger.error( "There are more ( " + count + " ) items with the same ItemName"  );
                    // ??? implement GUI: ask for delete all 
                    return;
                }
            }
            logger.fatal( "Delta Editor DELETE: " + deleteItemSql );
            if( stmt.execute( deleteItemSql ) )
            {
                logger.fatal( "DELETE ERROR : " + deleteItemSql );
            }
            stmt.close();
            // Update ( because GUI must changed )
            buildItemNameVector();
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );             
        }        
        return;
    }    
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------    
    public void renameItem( int iid, String newItemName, String oldItemName )
    {
        String renameItemSql = "UPDATE " + tableITEM + " SET \"ItemName\" = '" + newItemName + "' WHERE \"ItemName\" = '" + oldItemName + "' AND \"IID\" = " + iid;                  
        try
        {
            java.sql.Statement stmt = con.createStatement();
            logger.fatal( "Delta Editor RENAME: " + renameItemSql );
            if( stmt.execute( renameItemSql ) )
            {
                logger.fatal( "RENAME ERROR : " + renameItemSql );
            }
            stmt.close();
            // Update ( because GUI must changed )
            buildItemNameVector();
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );             
        }        
        return;
    }    
    //--------------------------------------------------------------------------
    //--------------------------------------------------------------------------    
    public void insertCopyNewItem( String itemName, int iid, int nameId )
    {
        try
        {
            int newIID = insertNewItem( itemName, nameId );            
            // copy item data                             
            String copySql = "SELECT \"CID\", \"Modifier\", \"CS\", \"X\", \"TXT\", \"Notes\", \"SEQ\" FROM " + tableDESCR + " WHERE \"IID\"=" + iid;
            java.sql.Statement stmt = con.createStatement();
            java.sql.ResultSet rs = stmt.executeQuery( copySql );
            java.util.Vector v = new java.util.Vector();            
            while( rs.next() )
            {
                v.add( new CharacterStateDescriptionEntry(   newIID,            // IID
                                                             rs.getInt(    1 ), // CID 
                                                             rs.getString( 2 ), // Modifier
                                                             rs.getString( 3 ), // CS
                                                             rs.getDouble( 4 ), // X
                                                             rs.getString( 5 ), // TXT
                                                             rs.getString( 6 ), // Notes
                                                             rs.getInt(    7 )  // SEQ
                                                             // ??? ,""     // characterType                                                 
                                                           ) 
                        );
            }
            rs.close();
            stmt.close();
            java.util.ListIterator iter = v.listIterator();            
            while( iter.hasNext() )
            {
                CharacterStateDescriptionEntry csde = (CharacterStateDescriptionEntry) iter.next();
                inserCharacterStateDescriptionEntry( csde );
            }
            // Update ( because GUI must changed )
            buildItemNameVector();
        }
        catch( java.sql.SQLException ex )
        {
            logger.fatal( ex.getMessage() );             
        }                
    }
    //--------------------------------------------------------------------------
    // Search ItemName
    //--------------------------------------------------------------------------
    public int getRowByItemName( String name )
    {
        java.util.ListIterator iter = itemNameVector.listIterator();
        
        while( iter.hasNext() )
        {
            ItemElement item = ( ItemElement ) iter.next();
            
            if( item.getItemName().startsWith( name ) )
            {
                // System.out.println( "Found: " + item.getItemName() );
                return iter.previousIndex();
            }           
        }
        return -1;
    }

    //--------------------------------------------------------------------------
    //
    //--------------------------------------------------------------------------
    public void setAcceptedNamesFlag( boolean flag)
    {
        showAcceptedNamesFlag = flag;
        buildItemNameVector();
    }
    //--------------------------------------------------------------------------
    //
    //--------------------------------------------------------------------------
    public void buildNaviKeyFiles( String fileSuffix, String sqlPreSelect, String excludeCharacters )
    {        
        buildNaviKeyStrings( sqlPreSelect, excludeCharacters );
        
        // write data to files
        try
        {
            java.io.OutputStreamWriter osw;

            java.io.FileOutputStream specsFile = new java.io.FileOutputStream( "specs." + fileSuffix );
            osw = new java.io.OutputStreamWriter( specsFile, "UTF-8" );
            osw.write( specs );
            osw.close();       
            specsFile.close();
            //------------------
            // write chars file
            //------------------            
            java.io.FileOutputStream charsFile = new java.io.FileOutputStream( "chars." + fileSuffix );
            osw = new java.io.OutputStreamWriter( charsFile, "UTF-8" );
            osw.write( chars );
            osw.close();       
            charsFile.close();

            //------------------
            // write items file
            //------------------            
            java.io.FileOutputStream itemsFile = new java.io.FileOutputStream( "items." + fileSuffix );
            osw = new java.io.OutputStreamWriter( itemsFile, "UTF-8" );
            osw.write( items );
            osw.close();       
            itemsFile.close();
        }
        catch( java.io.IOException ex )
        {
            logger.error( ex.getLocalizedMessage() );
        }
    }
    public String getItemsString()
    {
        return items;
    }
    public String getCharsString()
    {
        return chars;
    }
    public String getSpecsString()
    {
        return specs;
    }
    public void buildNaviKeyStrings()
    {
        buildNaviKeyStrings( null, null );
    }
    public void buildNaviKeyStrings( String preSelect, String excludeCharacters )
    {
        String preSelectWhere = null;
        String excludeWhere   = null;
        if( excludeCharacters == null || excludeCharacters.trim().length() <= 2 )
        {
            excludeWhere = "";
        }    
        else
        {
            excludeWhere = " WHERE \"CID\" NOT IN " +  excludeCharacters;
        }
        if( preSelect == null )
        {
            preSelectWhere = "";
        }
        else 
        {
            // TEST preselect: Sonora Desert
            // String pre = "SELECT DISTINCT \"IID\" FROM \"LIASlight_DESCR\" WHERE ( (\"CID\" = 13 AND \"CS\" IN ( 20 ) ) )";
            // preSelectWhere = "WHERE \"IID\" IN ( " + pre + " )";
        
            preSelectWhere = "WHERE \"IID\" IN ( " + preSelect + " )";
            // logger.info( )
        }
        
        // build specs
            StringBuffer specs = null;
            StringBuffer chars = null;
            StringBuffer items = null;
            String NEW_LINE = "\n";

            int DATA_BUFFER_SIZE            = 40000;
            int NUMBER_OF_CHARACTERS        = 71;
            int MAXIMUM_NUMBER_OF_STATES    = 837;
            int MAXIMUM_NUMBER_OF_ITEMS     = 6725;
            int MANDATORY_CHARACTERS        = 1;
            
            StringBuffer CHARACTER_TYPES        = new StringBuffer();
            StringBuffer NUMBERS_OF_STATES      = new StringBuffer();
            StringBuffer DEPENDENT_CHARACTERS   = new StringBuffer();
            StringBuffer CHARACTER_LIST         = new StringBuffer();
            StringBuffer ITEM_DESCRIPTIONS      = new StringBuffer();
            try
            {
                java.sql.Statement stmt = con.createStatement();
                
                String sql = null;
                java.util.Vector v = null;
                java.sql.ResultSet rs = null;
                // set NUMBER_OF_CHARACTERS
                logger.info( "set NUMBER_OF_CHARACTERS" );
                sql = "SELECT COUNT( * ) FROM " + tableCHAR + excludeWhere;
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                while( rs.next() )
                {
                    NUMBER_OF_CHARACTERS = rs.getInt( 1 );
                }
                rs.close();
                
                // set MAXIMUM_NUMBER_OF_ITEMS
                logger.info( "setup MAXIMUM_NUMBER_OF_ITEMS" );
                sql = "SELECT COUNT( * ) FROM " + tableITEM;
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                while( rs.next() )
                {
                    MAXIMUM_NUMBER_OF_ITEMS = rs.getInt( 1 );
                }
                rs.close();
                
                // set MAXIMUM_NUMBER_OF_STATES
                logger.info( "setup MAXIMUM_NUMBER_OF_STATES" );
                sql = "SELECT COUNT( * ) FROM " + tableCS +  excludeWhere;
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                while( rs.next() )
                {
                    MAXIMUM_NUMBER_OF_STATES = rs.getInt( 1 );
                }
                rs.close();
                
                // set CHARACTER TYPES
                logger.info( "setup CHARACTER TYPES" );                                
                sql = "SELECT \"CID\", \"Type\" FROM " + tableCHAR  + excludeWhere;
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                while( rs.next() )
                {
                    CHARACTER_TYPES.append( rs.getString( 1 ) );
                    CHARACTER_TYPES.append( "," );
                    CHARACTER_TYPES.append( rs.getString( 2 ) );
                    CHARACTER_TYPES.append( " " );
                }
                rs.close();
                                
                // set NUMBERS_OF_STATES
                
                logger.info( "setup NUMBERS_OF_STATES" );
                sql = "SELECT \"CID\", COUNT( \"CS\" )  FROM " + tableCS + excludeWhere + " GROUP BY \"CID\" ORDER BY \"CID\" "; 
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                while( rs.next() )
                {
                    NUMBERS_OF_STATES.append( rs.getString( 1 ) );
                    NUMBERS_OF_STATES.append( "," );
                    NUMBERS_OF_STATES.append( rs.getString( 2 ) );
                    NUMBERS_OF_STATES.append( " " );
                }
                rs.close();
                
                
                // set DEPENDENT_CHARACTERS
                logger.info( "setup DEPENDENT_CHARACTERS" );
                sql = "SELECT \"CID\", \"CS\", \"InapplicableCID\" FROM " + tableDEP +  excludeWhere + "ORDER BY \"CID\", \"CS\" ";
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                while( rs.next() )
                {
                    DEPENDENT_CHARACTERS.append( rs.getString( 1 ) );
                    DEPENDENT_CHARACTERS.append( "," );
                    DEPENDENT_CHARACTERS.append( rs.getString( 2 ) );
                    DEPENDENT_CHARACTERS.append( ":" );
                    DEPENDENT_CHARACTERS.append( rs.getString( 3 ) );
                    DEPENDENT_CHARACTERS.append( " " );
                }
                rs.close();
                
                //----------------------
                // CHARACTER STATES LIST
                //----------------------
                logger.info( "read " + tableCHAR );
                sql = "SELECT \"CID\", \"CharName\",\"Unit\" FROM " + tableCHAR +  excludeWhere + " ORDER BY \"CID\" ";
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                java.util.LinkedHashMap lhm = new java.util.LinkedHashMap();
                
                FormatUnit fu = new FormatUnit();
                
                
                while( rs.next() )
                {
                    String index = rs.getString( 1 );
                    String cs =  "\n#" + index + ". " + rs.getString( 2 ); // + "/\n";
                    String unit  =  rs.getString( 3 );
                    if( unit != null )
                    {
                        // numerical state
                        
                        // use the first part of the String
                        int pos = unit.indexOf( ' ' );
                        if( pos > 0 )
                        {    
                            unit = unit.substring( 0, pos );
                            unit = fu.format( unit );
                        }
                        cs += " " + unit; //  + "/\n";
                    }
                    cs += "/\n";
                    lhm.put( index, cs );
                }
                rs.close();

                logger.info( "read " + tableCS );
                // ??? StateWording most times are "null" 
                // ??? sql = "SELECT \"CID\",\"CS\",\"StateWording\" FROM " + tableCS + " ORDER BY \"CID\",\"StateID\" ";
                sql = "SELECT \"CID\",\"CS\",\"CharStateName\" FROM " + tableCS + excludeWhere + " ORDER BY \"CID\",\"StateID\" ";
                logger.info( sql );
                rs = stmt.executeQuery( sql );

                String lastIndex = "";
                
                while( rs.next() )
                {
                    String index = rs.getString( 1 );
                    String cs    = rs.getString( 2 );
                    String state = rs.getString( 3 );
                                        
                    if( index.compareTo( lastIndex) != 0 )
                    {
                        lastIndex = index;
                        CHARACTER_LIST.append( lhm.get( index ) );                        
                    }
                    try
                    {
                        java.lang.Integer.parseInt( cs );
                    }
                    catch( java.lang.NumberFormatException nfex )
                    {
                        // don't add this CharacterState (TE,V,U,-,...)
                        continue;
                    }
                    
                    CHARACTER_LIST.append( " " + cs + ". " + state +"/\n" );
                }
                rs.close();
                //--------------------
                // ITEM DESCRIPTIONS
                //--------------------
    
                logger.info( "read " + tableITEM );
                sql = "SELECT \"IID\",\"ItemName\", \"NameID\" FROM " + tableITEM + " ORDER BY \"IID\" ";
                logger.info( sql );
                rs = stmt.executeQuery( sql );
                lhm = new java.util.LinkedHashMap();
                
                while( rs.next() )
                {
                    String cs = "";
                    String name  = rs.getString( 2 );
                    String index = rs.getString( 1 );
                    // Don't show IID
                    // String cs =  "\n#" + index + ". " + name + "/\n";
                    if( diversityAcceptedNames != null )
                    {
                        // Build date with AcceptedNames                       
                        int id = rs.getInt( 3 );                        
                        String an = diversityAcceptedNames.getAcceptedName( id );                        
                        // System.out.println( cs + " --> " + an );                        
                        // JAVA 1.5 if( ! an.contentEquals( "" ) )
                        if( ! an.equalsIgnoreCase( "" ) )
                        {
                            name = an;
                        }
                        cs =  "\n#" + name + "/\n";
                    }
                    else
                    {
                        cs =  "\n#" + name + "/\n";                        
                    }    
                    lhm.put( index, cs );
                }
                rs.close();

                // For each item create the character state description
                logger.info( "read " + tableDESCR );
                sql = "SELECT \"IID\", \"CID\", \"Modifier\", \"CS\", \"X\", \"TXT\", \"Notes\", \"SEQ\" FROM " + tableDESCR + " ";
                
                String sqlOrderBy = " ORDER BY \"IID\", \"CID\" ";
                
                // add Where for preSelect
                
                String exclude = "";
                if( excludeWhere.compareTo( "" ) != 0 )
                {
                    exclude = " AND " + excludeWhere.substring( excludeWhere.indexOf( '"') -1 );
                }
                
                sql += preSelectWhere + exclude + sqlOrderBy;
                
                logger.info( "SQL " + sql );
                // get data
                rs  = stmt.executeQuery( sql );
                CharacterStateDescription csd = null;
                lastIndex = "";                
                int lastCid   = java.lang.Integer.MAX_VALUE;
                
                boolean lastStringFlag = false;
                
                while( rs.next() )
                {
                    lastStringFlag = true;
                    
                    int cid = rs.getInt( 2 );
                    
                    if( lastIndex.compareTo( rs.getString( 1 ) ) != 0 )
                    {                        
                        // Item ID Changed: write data
                        lastIndex = rs.getString( 1 );
                        if( csd != null )
                        {
                            ITEM_DESCRIPTIONS.append( csd.getItemDeltaFormatedString() );                            
                        }
                        ITEM_DESCRIPTIONS.append( lhm.get( lastIndex ) + " " );                        
                        // start with next item
                        csd = new CharacterStateDescription();                        
                    }
                    else if( lastCid != cid )
                    {
                        // character state changed
                        ITEM_DESCRIPTIONS.append( csd.getItemDeltaFormatedString() );
                        // start with next item
                        lastCid = cid;
                        csd = new CharacterStateDescription();                        
                    }
                    
                    csd.add( new CharacterStateDescriptionEntry( rs.getInt(    1 ), // IID
                                                                 rs.getInt(    2 ), // CID 
                                                                 rs.getString( 3 ), // Modifier
                                                                 rs.getString( 4 ), // CS
                                                                 rs.getDouble( 5 ), // X
                                                                 rs.getString( 6 ), // TXT
                                                                 rs.getString( 7 ), // Notes
                                                                 rs.getInt(    8 )  // SEQ
                                                           ) 
                        );
                }
                // write last String
                if( lastStringFlag == true )
                {
                    ITEM_DESCRIPTIONS.append( csd.getItemDeltaFormatedString() );
                }
                rs.close();
                stmt.close();            
            }
            catch( java.sql.SQLException ex )
            {
                logger.fatal( ex.getMessage() );             
            }                
            //------------------
            // write specs file
            //------------------
            
            specs = new StringBuffer();

            specs.append( "*SHOW ~ Dataset specifications." );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*DATA BUFFER SIZE " + DATA_BUFFER_SIZE  );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*NUMBER OF CHARACTERS " + NUMBER_OF_CHARACTERS );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*MAXIMUM NUMBER OF STATES " + MAXIMUM_NUMBER_OF_STATES );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*MAXIMUM NUMBER OF ITEMS " + MAXIMUM_NUMBER_OF_ITEMS );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*CHARACTER TYPES " + CHARACTER_TYPES.toString() );
            // specs.append( "*CHARACTER TYPES 10-11,TE 25,RN 36,RN 37-38,OM 46-47,RN 53,IN 55-56,RN 59,IN" );
            // specs.append( "  60,OM 64,RN 67,TE 71,TE" );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*NUMBERS OF STATES " + NUMBERS_OF_STATES.toString() );            
            // specs.append( "*NUMBERS OF STATES 1,5 2,62 3,3 4,64 6,5 7,837 8,145 9,23 12,9 13,29 14,8 15,9" );
            // specs.append( "  16,11 17,5 18,58 26,4 27,8 28,5 29,58 32,6 35,6 37,5 38,7 39,58 42,58 43,4" );
            // specs.append( "  44,58 48,58 49,5 52,6 54,43 60,8 61,5 63,43 66,607 70,3" );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*IMPLICIT VALUES" );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*DEPENDENT CHARACTERS " + DEPENDENT_CHARACTERS.toString() );
            // specs.append( "*DEPENDENT CHARACTERS 16,1/2/3/4/9/10/11:29-30 19,1:20 21,1:22-24 30,1:31-32" );
            // specs.append( "  31,1:32 34,1:35-61 35,3/4:38-41 40,1:41 51,1:52 57,1:58-59 65,1:66 68,1:69-70" );
            specs.append( NEW_LINE );
            specs.append( "" );
            specs.append( NEW_LINE );
            specs.append( "*MANDATORY CHARACTERS " + MANDATORY_CHARACTERS );
            specs.append( NEW_LINE );
                        
            this.specs = specs.toString();
            
            //-----------------------
            // chars
            //-----------------------
            chars = new StringBuffer();
            
            chars.append( "*SHOW ~ Character list. " );
            chars.append( NEW_LINE );
            chars.append( "*CHARACTER LIST " );
            chars.append( CHARACTER_LIST.toString() );
            chars.append( NEW_LINE );
                
            chars.append( NEW_LINE );
            
            this.chars = chars.toString();
            
            //----------------------
            // items
            //----------------------
            items = new StringBuffer();
            items.append( "*SHOW ~ Item descriptions" );
            items.append( NEW_LINE );
            items.append( "*ITEM DESCRIPTIONS" );
            items.append( NEW_LINE );
            items.append( ITEM_DESCRIPTIONS.toString() );
/*
            SELECT "IID","ItemName"  FROM "BIOTAlichendescr_ITEM";
 *
            "# Abrothallus microspermus Tul./"
            "1,2 2,61 3,2 4,63 5,2 6,2 7,1 8,1 9,23 12,1/5 13,28 14,5 16,1 34,1/2 35,1 49,5"
            "50,2 51,1 68,1<assumed>"
            
            # Acarospora cervina Massal./
            1,2 2,61 3,2 4,63 5,2 6,2 7,6 8,3 9,8 11<W95> 12,5/6 13,7/29 14,4 15,7 16,4
            18,6<light to dark>/41/45/46/51<often with pruinose surface and with whitish
            margin of squamules> 19,1/2 20,1 21,1 25,0.5-5 27,6/7 34,1/2 35,1
            36,1.5(-2.5)<max> 40,1 49,2 50,2 51,1<assumed> 55,4-8 56,1.5-3 57,1 60,1 61,1
            65,1 68,2 69,1 70,2
*/
            items.append( NEW_LINE );
            
            this.items = items.toString();            
    }
    //--------------------------------------------------------------------------
    //
    //--------------------------------------------------------------------------
    public static void main(String args[]) 
    {
/*
        <resource>Diversity Taxon Names Fungi</resource>
        <description>Info</description>
        <driver>net.sourceforge.jtds.jdbc.Driver</driver>
        <accesspoint>jdbc:jtds:sqlserver://141.84.65.123:5432/DiversityTaxonNames_Fungi</accesspoint>
        <user>www</user>
 */
        if( args.length != 6 )
        {
            System.out.println( "usage: DeltaEditorData url user pwd names_url names_user names_pwd");
            return;
        }
        
        try
        {
            java.sql.Connection con = null;
            
            java.sql.Connection names_con = null;
            // 
            
            String db_url   = args[ 0 ];
            String db_user  = args[ 1 ];
            String db_pwd   = args[ 2 ];
            
            // diversity taxon names lookup AcceptedNames
            
            String db_names_url     = args[ 3 ];
            String db_names_user    = args[ 4 ];
            String db_names_pwd     = args[ 5 ];
            
            System.out.println( "url  :" + db_url );
            System.out.println( "user :" + db_user );
            System.out.println( "names url  :" + db_names_url );
            System.out.println( "names user :" + db_names_user );
            System.out.println( "Build files ...");

            
            // load driver for postgres and MS-SQL and get DB-Connection's'
            try
            {   // PostgreSql
                Class.forName( "org.postgresql.Driver" );
                System.out.println( "Load PostgreSQL Driver..." );
                // MS-Sql-Server
                Class.forName( "net.sourceforge.jtds.jdbc.Driver" );
                System.out.println( "Load MS-SQL-Server Driver..." );
            }
            catch( java.lang.ClassNotFoundException ee )
            {
                System.out.println( "ClassNotFoundException: " );
                System.out.println( ee.getMessage() );
            }
            try
            {
                con = java.sql.DriverManager.getConnection( db_url, db_user, db_pwd );             
                names_con = java.sql.DriverManager.getConnection( db_names_url, db_names_user, db_names_pwd );
            }            
            catch( java.sql.SQLException ex )
            {
                System.out.println( ex.getMessage() );             
            }
            
            DeltaEditorData ded = new DeltaEditorData( con, names_con, "LIASlight" );            
            
            ded.buildNaviKeyFiles( "LIASlight", null, null );
            
            // build all subsets
            try
            {
                String databaseName = "LIASlight";
        
                javax.xml.parsers.DocumentBuilderFactory builderFactory = javax.xml.parsers.DocumentBuilderFactory.newInstance();
                javax.xml.parsers.DocumentBuilder builder = builderFactory.newDocumentBuilder();
                org.w3c.dom.Document document = builder.parse( new org.xml.sax.InputSource( "delta_pre_selection_" + databaseName + ".xml" ) );
                Delta_pre_selectionScanner scanner = new Delta_pre_selectionScanner( databaseName, document);
                scanner.visitDocument();          

                String [] names = scanner.getPreSelectionNames();

                System.out.println( "--- all subsets ---" );        
                for( int i = 0; i < names.length; i++ )
                {    
                    String subsetName = names[ i ];
                    System.out.println( "--- get Sql for " + subsetName );
                    String sqlPreSelection = scanner.getSql( subsetName );
                    String excludeCharacters = scanner.getExclude( subsetName );
                    System.out.println( sqlPreSelection );
                    System.out.println( excludeCharacters );
                    String fileSuffix = subsetName.replace( ' ', '_' );
                    System.out.println( "--- build " + databaseName + "." + fileSuffix );
                    
                    ded.buildNaviKeyFiles( fileSuffix, sqlPreSelection, excludeCharacters );            
                    
                }
            }
            catch( org.xml.sax.SAXException ex )
            {
                System.out.println( ex.getMessage() );             
            }
            catch( javax.xml.parsers.ParserConfigurationException ex )
            {
                System.out.println( ex.getMessage() );             
            }
            catch( java.io.IOException ex )
            {
                System.out.println( ex.getMessage() );             
            }
        }
        /*
        catch( java.sql.SQLException ex )
        {
            ex.printStackTrace();
        }
         */
        finally
        {
            
        }                
    }
    
}
