/*
 * DiversityAcceptedNames.java
 *
 * Created on July 6, 2006, 3:35 PM
 *
 */

package net.metadiversity.diversity.navigator.util;

import org.apache.log4j.Logger;
import java.lang.Integer;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
/**
 *
 * @author didi
 */
public class DiversityAcceptedNames 
{
    static Logger logger = Logger.getLogger( DiversityAcceptedNames.class.getName() ); 

    private String url;
    private String user;
    private String pwd;

    Connection conn;

    private static String MS_SQL_DRIVER_CLASS     = "net.sourceforge.jtds.jdbc.Driver";
    private static String POSTGRESQL_DRIVER_CLASS = "org.postgresql.Driver";

    // JAVA 1.5 private HashMap< Integer, String > cache = null; 
    private HashMap cache = null;
    private boolean useCache = true;
    
    /** Creates a new instance of DiversityAcceptedNames */
    public DiversityAcceptedNames( String url, String user, String pwd ) 
    {
        this.url  = url;
        this.user = user;
        this.pwd  = pwd;
        
        initConnection();
        if( useCache )
        {
            initCache();
        }
        init();
    }

    /** Creates a new instance of DiversityAcceptedNames */
    public DiversityAcceptedNames( java.sql.Connection conn ) 
    {
        this.conn = conn;
        
        init();
        if( useCache )
        {
            initCache();
        }
    }
    
    public void init()
    {
        
    }
    public void initCache()
    {
        // JAVA 1.5 cache = new java.util.HashMap< Integer, String >();
        cache = new java.util.HashMap();
        
        logger.info( "Build accepted names cache data ..." );
        
        String acceptedName = "";
        String sql = "SELECT \"AcceptedName\", \"NameID\" FROM \"AcceptedNames\"";
        java.sql.Statement statement;
        try 
        {
            statement = conn.createStatement();
            java.sql.ResultSet res = statement.executeQuery( sql );
            
            while( res.next() )
            {
                acceptedName = res.getString( 1 );
                if( acceptedName == null ) acceptedName = "";
                if( acceptedName.equalsIgnoreCase( "null" ) ) acceptedName = "";
                Integer id = new Integer( res.getInt( 2 ) );
                logger.debug( id + ": " + acceptedName );
                cache.put( id, acceptedName );
            }
            // System.out.println( " OK" );
            logger.info( " OK" );
        } 
        catch (SQLException ex) 
        {
            ex.printStackTrace();
        }        
    }
    public void initConnection()
    {
        //-------------------------------
        // Get connection to DiversityTaxonNamesCache
        //-------------------------------
        try
        {
            logger.info( "Load "  + POSTGRESQL_DRIVER_CLASS );
            Class.forName( POSTGRESQL_DRIVER_CLASS );
        }
        catch( java.lang.ClassNotFoundException ee )
        {
            System.out.println( "ClassNotFoundException: " );
            System.out.println( ee.getMessage() );
            
            System.exit( 3 );
        }
        try
        {
            conn = java.sql.DriverManager.getConnection( url, user, pwd );
        }            
        catch( java.sql.SQLException ex )
        {
            System.out.println( ex.getMessage() );             
            System.exit( 4 );
        }        
 
    }
    
    public String getAcceptedName( int id )
    {
        String acceptedName = "";
        if( useCache )
        {
            // JAVA 1.5 acceptedName = cache.get( new Integer( id ) );
            acceptedName = (String) cache.get( new Integer( id ) );
            
            if( acceptedName == null ) acceptedName = "";
            return acceptedName;
        }
        String sql = "SELECT \"AcceptedName\" FROM \"AcceptedNames\" WHERE \"NameID\" = " + id;
        java.sql.Statement statement;
        try 
        {
            statement = conn.createStatement();
            java.sql.ResultSet res = statement.executeQuery( sql );
            
            if( res.next() )
            {
                acceptedName = res.getString( 1 );
            }
        } 
        catch (SQLException ex) 
        {
            ex.printStackTrace();
        }

        return acceptedName;        
    }
     /**
     * @param args the command line arguments
     */
    public static void main( String[] args ) 
    {
        String url  = "jdbc:postgresql://141.84.65.114:5432/AcceptedNamesCache";
        String user = "xxxx";
        String pwd  = "********";
        
        DiversityAcceptedNames dan = new DiversityAcceptedNames( url, user, pwd );
        
        for(int i = 1000000; i < 1000999; i += 25 )
        {
            System.out.println( i + " " + dan.getAcceptedName( i ) );
        }
        System.exit( 0 );
    }
}
